from typing import Any, Dict, cast

import httpx

from ._api_commons import assert_response_status
from ..client import Client
from ..dto import RunDto

content_type_merge_patch = 'application/merge-patch+json'


def create_run(*, client: Client, project_key: str, run: RunDto) -> RunDto:

    url = "{}/projects/{projectKey}/runs".format(client.base_url, projectKey=project_key)

    headers: Dict[str, Any] = client.get_headers()

    json_body = run.to_dict_without_none_values()

    response = httpx.post(url=url, headers=headers, json=json_body)

    assert_response_status(response)

    return RunDto.from_dict(cast(Dict[str, Any], response.json()))


def partial_update_run(*, client: Client, project_key: str, run_key: int, run: RunDto) -> None:

    url = "{}/projects/{projectKey}/runs/{runKey}".format(
        client.base_url, projectKey=project_key, runKey=run_key
    )

    headers: Dict[str, Any] = client.get_headers()
    headers["content-type"] = content_type_merge_patch

    json_body = run.to_dict_without_none_values()

    response = httpx.patch(url=url, headers=headers, json=json_body)

    assert_response_status(response)


def update_run_parameters(*, client: Client, project_key: str, run_key: int, parameters: Dict[str, Any]) -> None:

    url = "{}/projects/{projectKey}/runs/{runKey}/parameters".format(
        client.base_url, projectKey=project_key, runKey=run_key
    )

    headers: Dict[str, Any] = client.get_headers()
    headers["content-type"] = content_type_merge_patch

    response = httpx.patch(url=url, headers=headers, json=parameters)

    assert_response_status(response)


def update_run_metrics(*, client: Client, project_key: str, run_key: int, metrics: Dict[str, Any]) -> None:

    url = "{}/projects/{projectKey}/runs/{runKey}/metrics".format(
        client.base_url, projectKey=project_key, runKey=run_key
    )

    headers: Dict[str, Any] = client.get_headers()
    headers["content-type"] = content_type_merge_patch

    response = httpx.patch(url=url, headers=headers, json=metrics)

    assert_response_status(response)



from git import Repo, InvalidGitRepositoryError, NoSuchPathError
from typing import Optional
from .model.git import Git


def get_git_metadata(remote_name: str = 'origin') -> Optional[Git]:
    try:
        repo = Repo(search_parent_directories=True)
    except (InvalidGitRepositoryError, NoSuchPathError):
        return None

    git_metadata = Git(
        commit_time=repo.head.object.committed_datetime,
        commit_hash=repo.head.object.hexsha,
        is_dirty=repo.is_dirty(),
        repository_uri=next(repo.remote(remote_name).urls)
    )

    return git_metadata



import datetime
from dataclasses import dataclass
from dataclasses_json import DataClassJsonMixin, config, dataclass_json, LetterCase, Undefined

from .helper import datetime_field, ExtendedDtoSerializer


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GitDto(ExtendedDtoSerializer, DataClassJsonMixin):
    dataclass_json_config = config(
        letter_case=LetterCase.CAMEL,
        undefined=Undefined.EXCLUDE
    )['dataclasses_json']

    is_dirty: bool
    commit_time: datetime.datetime = datetime_field()
    commit_hash: str = None
    repository_uri: str = None



from dataclasses import dataclass
from dataclasses_json import DataClassJsonMixin, config, LetterCase, Undefined
from datetime import datetime
from enum import Enum
from typing import List, Optional

from .helper import datetime_field, ExtendedDtoSerializer


class ExperimentStatusDto(str, Enum):
    TODO = "TODO"
    IN_PROGRESS = "IN_PROGRESS"
    COMPLETED = "COMPLETED"


@dataclass
class ExperimentDto(ExtendedDtoSerializer, DataClassJsonMixin):
    dataclass_json_config = config(
        letter_case=LetterCase.CAMEL,
        undefined=Undefined.EXCLUDE
    )['dataclasses_json']

    created_at: Optional[datetime] = datetime_field()
    key: Optional[str] = None
    name: Optional[str] = None
    status: Optional[ExperimentStatusDto] = None
    tags: Optional[List[str]] = None



from .model import Artifact, ArtifactRef, Git, Run, RunStatus
from ._api_client.dto import ArtifactDto, ArtifactRefDto, GitDto, ExperimentRefDto, RunDto, StatusDto

from typing import List, Optional


def dto_to_artifact(artifact_dto: ArtifactDto) -> Artifact:
    return Artifact(
        created_at=artifact_dto.created_at,
        created_by=artifact_dto.created_at,
        files=artifact_dto.files,
        metadata=artifact_dto.metadata,
        model=artifact_dto.model,
        name=artifact_dto.name,
        run_key=artifact_dto.run_key,
        run_name=artifact_dto.run_name,
        type=artifact_dto.type,
        updated_at=artifact_dto.updated_at,
        version=artifact_dto.version,
    )


def run_to_dto(run: Run, experiment_key: Optional[str], used_artifacts: Optional[List[ArtifactRef]]) -> RunDto:
    return RunDto(
        created_at=None,
        created_by=None,
        end_time=None,
        experiment_refs=None if experiment_key is None else [ExperimentRefDto(experiment_key)],
        git=None if run.git is None else git_to_dto(run.git),
        key=None,
        metrics=None,
        name=run.name,
        parameters=None,
        start_time=None,
        status=StatusDto(run.status.name),
        used_artifacts=__map_artifact_refs(used_artifacts)
    )


def dto_to_run(run_dto: RunDto) -> Run:
    return Run(
        name=run_dto.name,
        status=RunStatus[run_dto.status.name],
        parameters={} if run_dto.parameters is None else run_dto.parameters,
        metrics={} if run_dto.metrics is None else run_dto.metrics,
        start_time=run_dto.start_time,
        end_time=run_dto.end_time,
        key=run_dto.key,
    )


def git_to_dto(git: Git) -> GitDto:
    return GitDto(
        commit_time=git.commit_time,
        commit_hash=git.commit_hash,
        is_dirty=git.is_dirty,
        repository_uri=git.repository_uri
    )


def __map_artifact_refs(artifacts: List[ArtifactRef] = None):
    return list(map(lambda a: ArtifactRefDto(name=a.name, version=a.version), artifacts)) if artifacts else None



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, cast, Optional
from httpx import Response


@dataclass
class HttpError:
    code: int
    message: Optional[str] = None

    def to_dict(self) -> Dict[str, Any]:
        code = self.code
        message = self.message

        return {
            "code": code,
            "message": message,
        }

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> HttpError:
        code = d["code"]
        message = d["message"]

        return HttpError(code=code, message=message)

    @staticmethod
    def from_response(response: Response):
        if response.text:
            try:
                response_body = cast(Dict[str, Any], response.json())
                error = HttpError.from_dict(response_body)
            except (KeyError, ValueError):
                error = HttpError(code=response.status_code)
        else:
            error = HttpError(code=response.status_code)

        return error


class MLAideError(Exception):
    http_error: HttpError
    """Base class for exceptions in ML Aide.
    
    Attributes:
        http_error: Details about the HTTP error, if the error was raised for an invalid HTTP response.
    """

    def __init__(self, http_error: HttpError = None):
        self.http_error = http_error


class InputError(MLAideError):
    """Exception raised for errors in the input."""


class NotFoundError(MLAideError):
    """Exception raised for errors where the requested object does not exist."""


class InvalidAuthorizationError(MLAideError):
    """Exception raised for invalid authorization (api key)."""
    pass


class NotAuthorizedError(MLAideError):
    """Exception raised for not sufficient permissions."""
    pass


class ServerError(MLAideError):
    """Exception raised for ML Aide server errors."""
    pass



from mlaide.error import NotFoundError
from . import _model_deser
from ._api_client import Client
from ._api_client.api import run_api, artifact_api, experiment_api
from ._api_client.dto import ArtifactDto, ExperimentDto, ExperimentStatusDto, RunDto, StatusDto
from .model import Artifact, ArtifactRef, Git, Run, RunStatus
from .mapper import dto_to_run, run_to_dto, dto_to_artifact

from datetime import datetime
from typing import Dict, List, Optional, Union
from io import BytesIO
from pathlib import Path
from os.path import relpath


class ActiveRun(object):
    """This class provides access to runs that are stored in ML Aide"""

    __api_client: Client
    __run: Run
    __project_key: str

    def __init__(self,
                 api_client: Client,
                 project_key: str,
                 run_name: str,
                 git: Optional[Git] = None,
                 experiment_key: Optional[str] = None,
                 used_artifacts: Optional[List[ArtifactRef]] = None,
                 auto_create_experiment: bool = True):
        self.__api_client = api_client
        self.__project_key = project_key

        if auto_create_experiment and experiment_key is not None:
            self.__create_experiment_if_not_present(project_key, experiment_key)

        self.__run = self.__create_new_run(experiment_key, run_name, git, used_artifacts)

    def __create_new_run(self,
                         experiment_key: str,
                         run_name: str,
                         git: Optional[Git] = None,
                         used_artifacts: Optional[List[ArtifactRef]] = None) -> Run:
        run = Run(name=run_name, status=RunStatus.RUNNING, git=git)
        run_to_create = run_to_dto(run, experiment_key, used_artifacts)

        created_run: RunDto = run_api.create_run(
            client=self.__api_client,
            project_key=self.__project_key,
            run=run_to_create
        )

        return dto_to_run(created_run)

    def __create_experiment_if_not_present(self, project_key: str, experiment_key: str):
        experiment = experiment_api.get_experiment(client=self.__api_client,
                                                   project_key=project_key,
                                                   experiment_key=experiment_key)

        if experiment is None:
            experiment = ExperimentDto(key=experiment_key, name=experiment_key, status=ExperimentStatusDto.IN_PROGRESS)
            experiment_api.create_experiment(client=self.__api_client, project_key=project_key, experiment=experiment)

    @property
    def run(self) -> Run:
        # Return a deep copy to avoid changing anything by the client
        return Run(
            start_time=self.__run.start_time,
            end_time=self.__run.end_time,
            status=self.__run.status,
            metrics=self.__run.metrics,
            parameters=self.__run.parameters,
            key=self.__run.key,
            name=self.__run.name
        )

    def log_metric(self, key: str, value) -> Run:
        """Logs a metric

        Arguments:
            key: The key of the metric.
            value: The value of the metric. The value can be any type that is JSON serializable.
        """
        self.__run.metrics[key] = value
        run_api.update_run_metrics(
            client=self.__api_client,
            project_key=self.__project_key,
            run_key=self.__run.key,
            metrics={key: value})
        return self.__run

    def log_metric_epoch(self, key: str, epoch: str, value) -> Run:
        """Logs a metric for an epoch

        Arguments:
            key: The key of the metric.
            epoch: The corresponding epoch.
            value: The value of the metric. The value can be any type that is JSON serializable.
        """
        if self.__run.metrics.get(key):
            new_dict = dict(self.__run.metrics.get(key))
            new_dict.update({epoch: value})
            self.__run.metrics[key] = new_dict
        else:
            self.__run.metrics[key] = {epoch: value}
        run_api.update_run_metrics(
            client=self.__api_client,
            project_key=self.__project_key,
            run_key=self.__run.key,
            metrics={key: self.__run.metrics[key]})
        return self.__run


    def log_parameter(self, key: str, value) -> Run:
        """Logs a parameter

        Arguments:
            key: The key of the parameter.
            value: The value of the parameter. The value must be a scalar value (e.g. string, int, float, ...).
        """
        self.__run.parameters[key] = value
        run_api.update_run_parameters(
            client=self.__api_client,
            project_key=self.__project_key,
            run_key=self.__run.key,
            parameters={key: value})
        return self.__run

    def log_model(self, model, model_name: str, metadata: Optional[Dict[str, str]] = None):
        """Creates a new artifact with type 'model'. The artifact will be registered as model.

        Arguments:
            model: The model. The model must be serializable.
            model_name: The name of the model. The name will be used as artifact filename.
            metadata: Some optional metadata that will be attached to the artifact.
        """
        artifact = self.create_artifact(name=model_name, artifact_type='model', metadata=metadata)

        _model_deser.serialize(
            model, 
            lambda serialized_model, filename: self.add_artifact_file(artifact=artifact, file=serialized_model, filename=filename))

        artifact_api.create_model(
            client=self.__api_client,
            project_key=self.__project_key,
            artifact_name=artifact.name,
            artifact_version=artifact.version)

    def get_or_create_artifact(self, name: str, artifact_type: str, metadata: Optional[Dict[str, str]]) -> Artifact:
        """Tries to get an existing artifact. If the artifact with the specified name does not exist a 
        new artifact will be registered.

        The artifact type and metadata will only be used for new artifacts. Existing artifacts will not be updated.

        Arguments:
            name: The name of the artifact.
            artifact_type: The artifact type.
            metadata: Some optional metadata that will be attached to the artifact.
        """
        try:
            return artifact_api.get_artifact(
                client=self.__api_client,
                project_key=self.__project_key,
                artifact_name=name,
                artifact_version=None)
        except NotFoundError:
            return self.create_artifact(name, artifact_type, metadata)

    def create_artifact(self, name: str, artifact_type: str, metadata: Optional[Dict[str, str]]) -> Artifact:
        """Creates a new artifact. If an artifact with the same name already exists, a new artifact with the
        next available version number will be registered.

        Arguments:
            name: The name of the artifact.
            artifact_type: The artifact type.
            metadata: Some optional metadata that will be attached to the artifact.
        """
        artifact_dto = ArtifactDto(name=name, type=artifact_type, metadata=metadata, run_key=self.__run.key)

        artifact_dto = artifact_api.create_artifact(
            client=self.__api_client,
            project_key=self.__project_key,
            artifact=artifact_dto)

        return dto_to_artifact(artifact_dto)

    def add_artifact_file(self, artifact: Artifact, file: Union[str, BytesIO], filename: str = None):
        """Add a file to an existing artifact. To add multiple file, specify a directory or invoke this function
        multiple times.

        Arguments:
            artifact: The artifact to which the file should be added.
            file: The file that should be added. This can be a io.BytesIO object or a string to a file or directory.
            filename: The filename. If the file is of type BytesIO the filename must be specified. If the file is a
                string, the original filename will be the default.
        """
        artifact_api.upload_file(
            client=self.__api_client,
            project_key=self.__project_key,
            artifact_name=artifact.name,
            artifact_version=artifact.version,
            filename=filename if filename is not None else ActiveRun.__extract_filename(file),
            file=ActiveRun.__normalize_file(file))

    @staticmethod
    def __extract_filename(file: Union[str, BytesIO]) -> str:
        if isinstance(file, str):
            return relpath(file)

        raise Exception('filename must be provided if provided file is of type io.BytesIO')

    @staticmethod
    def __normalize_file(file: Union[str, BytesIO]) -> BytesIO:
        if isinstance(file, str):  # Read the file behind the string/path
            if file.startswith('http://') or file.startswith('https://'):  # The file must be downloaded
                pass
                # TODO: Download file

            else:  # The file must be read from filesystem
                path = Path(file)

                if path.is_file():  # check if it is only a single file
                    file_bytes = path.read_bytes()
                    return BytesIO(file_bytes)

                elif path.is_dir():  # ... or is it a directory?
                    pass
                    # TODO: Read all files from directory

        else:  # The passed file is already of type BytesIO
            return file

    def set_completed_status(self) -> Run:
        """Sets the status of the current run as completed."""
        return self._set_status(RunStatus.COMPLETED)

    def set_failed_status(self) -> Run:
        """Sets the status of the current run as failed."""
        return self._set_status(RunStatus.FAILED)

    def _set_status(self, status: RunStatus) -> Run:
        self.__run.end_time = datetime.now()
        self.__run.status = status
        run_api.partial_update_run(
            client=self.__api_client,
            project_key=self.__project_key,
            run_key=self.__run.key,
            run=RunDto(
                status=StatusDto(status.name)
            )
        )
        return self.__run



import io
import cgi
from typing import Any, Tuple, cast, Dict, Optional

import httpx

from ._api_commons import assert_response_status
from ..client import Client
from ..dto import ArtifactDto


def create_model(*, client: Client, project_key: str, artifact_name: str, artifact_version: int) -> None:
    url = "{}/projects/{projectKey}/artifacts/{artifactName}/{artifactVersion}/model"\
        .format(client.base_url, projectKey=project_key, artifactName=artifact_name, artifactVersion=artifact_version)

    headers: Dict[str, Any] = client.get_headers()

    response = httpx.request(
        method="PUT",
        url=url,
        headers=headers,
    )

    assert_response_status(response)


def create_artifact(*, client: Client, project_key: str, artifact: ArtifactDto) -> ArtifactDto:
    url = "{}/projects/{projectKey}/artifacts"\
        .format(client.base_url, projectKey=project_key)

    headers: Dict[str, Any] = client.get_headers()

    response = httpx.request(
        method="POST",
        url=url,
        headers=headers,
        json=artifact.to_dict_without_none_values()
    )

    assert_response_status(response)

    return ArtifactDto.from_dict(response.json())


def upload_file(*, client: Client, project_key: str, artifact_name: str, artifact_version: int, filename: str, file: io.BytesIO):
    url = "{}/projects/{projectKey}/artifacts/{artifactName}/{artifactVersion}/files"\
        .format(client.base_url, projectKey=project_key, artifactName=artifact_name, artifactVersion=artifact_version)

    headers: Dict[str, Any] = client.get_headers()

    files = {'file': (filename, file)}

    response = httpx.request(
        method="POST",
        url=url,
        headers=headers,
        files=files
    )

    assert_response_status(response)


def get_artifact(*, client: Client,
                 project_key: str,
                 artifact_name: str,
                 artifact_version: Optional[int],
                 model_stage: str = None) -> ArtifactDto:
    version = artifact_version if artifact_version is not None else "latest"

    url = "{}/projects/{projectKey}/artifacts/{artifactName}/{artifactVersion}" \
        .format(client.base_url, projectKey=project_key, artifactName=artifact_name, artifactVersion=version)

    query_params = None if model_stage is None else {"model-stage": model_stage}

    headers: Dict[str, Any] = client.get_headers()

    response = httpx.request(
        method="GET",
        url=url,
        headers=headers,
        params=query_params
    )

    assert_response_status(response)

    return ArtifactDto.from_dict(cast(Dict[str, Any], response.json()))


def download_artifact(*,
                      client: Client,
                      project_key: str,
                      artifact_name: str,
                      artifact_version: int) -> Tuple[io.BytesIO, str]:

    url = "{}/projects/{projectKey}/artifacts/{artifactName}/{artifactVersion}/files" \
        .format(client.base_url, projectKey=project_key, artifactName=artifact_name, artifactVersion=artifact_version)

    headers: Dict[str, Any] = client.get_headers()
    headers["Accepts"] = "application/zip"

    response = httpx.request(
        method="GET",
        url=url,
        headers=headers
    )

    assert_response_status(response)

    content_disposition = response.headers.get("Content-Disposition")
    value, params = cgi.parse_header(content_disposition)
    filename = params.get("filename")
    return io.BytesIO(response.content), filename



import io
from typing import Any, Callable, Union
import cloudpickle
import tempfile
from pathlib import Path


def serialize(model: Any, save_callback: Callable[[Union[str, io.BytesIO], str], None]):
    for class_obj in model.__class__.__mro__:
        module_name = class_obj.__module__
        if not module_name:
            continue
        # elif module_name.startswith("torch"):
        #     model_type = "torch"
        #     bytestream, method = ensure_bytestream(model)
        #     break
        elif module_name.startswith("sklearn"):
            buffer = io.BytesIO()
            cloudpickle.dump(model, buffer)
            buffer.seek(0)

            save_callback(buffer, 'model.pkl')
            break
        # elif module_name.startswith("xgboost"):
        #     model_type = "xgboost"
        #     bytestream, method = ensure_bytestream(model)
        #     break
        elif module_name.startswith("keras.engine.training"):
            with tempfile.TemporaryDirectory() as tmpdirname:
                model.save(tmpdirname, save_format='tf')

                tmpdir_path = Path(tmpdirname)
                pathlist = tmpdir_path.rglob('*')
                for path in pathlist:
                    if (path.is_file()):
                        relative_path = str(path.relative_to(tmpdir_path))
                        save_callback(str(path), relative_path)

            break



from httpx import Response


class ApiResponseError(Exception):
    """ An exception raised when an unknown response occurs """

    def __init__(self, *, response: Response, error=None):
        super().__init__()
        self.response: Response = response
        self.error = error




from dataclasses import dataclass
from typing import Dict


@dataclass
class Client:
    """ A class for keeping track of data related to the API """

    base_url: str

    def get_headers(self) -> Dict[str, str]:
        """ Get headers to be used in all endpoints """
        return {}


@dataclass
class AuthenticatedClient(Client):
    """ A Client which has been authenticated for use on secured endpoints """

    api_key: str

    def get_headers(self) -> Dict[str, str]:
        """ Get headers to be used in authenticated endpoints """
        return {"x-api-key": f"{self.api_key}"}



import datetime
from enum import Enum
from typing import Dict, Optional
from dataclasses import dataclass, field
from . import Git


class RunStatus(str, Enum):
    RUNNING = 'RUNNING',
    FAILED = 'FAILED',
    COMPLETED = 'COMPLETED',


@dataclass
class Run(object):
    git: Optional[Git] = None
    key: Optional[int] = None
    name: Optional[str] = None
    start_time: datetime = None
    end_time: datetime = None
    status: Optional[RunStatus] = None
    metrics: Dict[str, any] = field(default_factory=dict)
    parameters: Dict[str, str] = field(default_factory=dict)
